require("table")
require("string")
local class = require("classy")

function table.copy(t)
  local u = { }
  for k, v in pairs(t) do u[k] = v end
  return setmetatable(u, getmetatable(t))
end

-- ++++++++++++++++++++++

InvItem = class('InvItem')

function InvItem:__init(o, r)
  -- print('InvItem:init', o)
  self.count = 0
  self.unit = ''
  self.text = ''
  self.tax = r.tax
  self.taxmode = r.taxmode
  self.amount = 0
  self.typ = 'any' -- e.g. time, lump, text
  self.discount = r.discount
  for key, val in pairs(o) do
    self[key] = val
  end
end

function InvItem:validate(key, list)
  -- is the item’s key value within the list?
  if not list then
    print(list, "is not a valid list but ", type(list))
    return false
  end
  for _, v in pairs(list) do
    if self[key] == v then
      return true
    end
  end
  print("value of '" .. key .. "' is not valid: '" .. self[key] .. "'" .. type(key))
  return false
end

function InvItem:sum()
  return self.count * self.amount * (100 - self.discount) / 100
end

function InvItem:taxsum()
  -- included or added VAT
  local sum = self:sum()
  if self.taxmode == 'brutto' then
    return sum - (sum / (1 + self.tax / 100))
  else
    return sum / 100 * self.tax
  end
end

-- ++++++++++++++++++++++

Invoice = class('Invoice')

function Invoice:__init(o)
  -- print('Invoice:init', o)
  self._zerosums = { [7] = 0, [19] = 0, [0] = 0, hours = 0, netto = 0, brutto = 0, discount = 0 }
  self.valid_tax = { 0, 7, 19 }
  self.valid_typ = { 'any', 'time', 'lump', 'text', 'porto'}
  self.sums = {}
  self.nettosums = true -- rechne mit Netto-Summen // taxmode?
  self.declaretax = true -- weise MwSt. aus
  self.tax = 19
  self.taxmode = 'brutto'
  self.discount = 0 -- Rabatt in Prozent
  self.printtitle = false
  self.printprice = not tex.modes['lieferschein'] -- Lieferschein: false
  self.printdiscounttitle = false
  self.printhoursum = false
  self.perhour = 50 -- Stundensatz
  self.discount = 0.0 -- Rabatt in Prozent
  self.alldiscounts = {}
  self.currency1 = "€"
  self.currency2 = "€" --"US\\$"
  self.exchangerate = 1/1.0707 -- currency2-Preis = currency1-Preis * exchangerate
  self.defaultunit = ""
  self.timeunit = "h"
  self.allunits = {} -- Liste der verwendeten Einheiten
  self.items = {} -- Posten
  self.language = "de"
  self.numberseparator = { default = ".", de = ",", en = "." }
  self.taxname = { default = "tax", de = "{MwSt․}", en = "VAT"}
  self.sumname = { default = "∑", de = "gesamt", en = "sum" }
  self.singlename = { default = "single", de = "einzeln", en = "single" }
  self.amountname = { default = "", de = "{Anz․}", en = "{Qnt․}" }
  self.textname = { default = "", de = "Bezeichnung", en = "title" }
  self.discountname = { default = "", de = "Rabatt", en = "discount" }
  self.bold = '\\bf '

  self.sums = table.copy(self._zerosums)
  for key, val in pairs(o) do
    self[key] = val
  end
end

function Invoice:addItem(o)
  -- print('Invoice:addItem', o)
  item = InvItem(o, self)
  if not item:validate("tax", self.valid_tax) then
    print("value of 'tax' is not valid: '" .. item.tax .. "'")
    return nil
  end
  if not item:validate("typ", self.valid_typ) then
    print("value of 'typ' is not valid: '" .. item.typ .. "'")
    return nil
  end
  if item.typ == 'lump' then
    item.count = 1
  elseif item.typ == 'porto' then
    item.count = 1
    -- tax: like most of other items...
  elseif item.typ == 'time' and o.amount == nil then
    item.amount = self.perhour
  end
  if item.discount > 0 then
    self.printdiscounttitle = true
  end
  table.insert(self.items, item)
  -- self.items[#self.items+1] = item
  return item
end

function Invoice:summarize()
  -- calculate sums
  self.sums = table.copy(self._zerosums)
  for _, item in pairs(self.items) do
    self.sums[item.tax] = self.sums[item.tax] + item:taxsum()
    self.sums.netto = self.sums.netto + item:sum()
    if self.taxmode == 'netto' then
        self.sums.brutto = self.sums.brutto + item:sum() + item:taxsum()
    else
        self.sums.brutto = self.sums.brutto + item:sum()
    end
    if (item.typ == "time") or (item.unit == self.timeunit) then
      self.sums.hours = self.sums.hours + item.count
    end
    self.allunits[item.unit] = true
    if item.discount then
        self.alldiscounts[item.discount] = true
        self.sums.discount = self.sums.discount + item:sum() * item.discount/100
    end
  end
  return self.sums
end

function Invoice:fmt_currency(amount, symbol)
  -- format number to currency
  if not symbol then
    symbol = self.currency1
  end
  local sep = self.numberseparator[self.language]
  return "" .. string.gsub(string.format('%.2f\\,%s', amount, symbol), "%.", sep)
end

function Invoice:fmt_percent(value, hideifempty)
  -- format number to percent
  if value == 0 and hideifempty then
      return ""
  end
  return "" .. string.format('%d\\,\\%%', value)
end

function Invoice:fmt_number(value, hideifempty)
  -- format number
  if value == 0 and hideifempty then
      return ""
  end
  local sep = self.numberseparator[self.language]

  if tonumber(value) == math.floor(value) then
    return "" .. string.format('%d', value)
  end
  return "" .. string.gsub(string.format('%0.2f', value), "%.", sep)
end

function Invoice:output()
  local sep = self.numberseparator[self.language]
  -- TODO: Unterscheidung brutto/netto, Rabatt; Einheit nur anzeigen, wenn sinnvoll
  local fullwidth = 0.75 + 0.75 + 7.5 + 1.5 + 1.75 + 2.25
  local countwidth = 0.75
  local unitwidth = 0.5
  local textwidth = 9.25
  local taxwidth = 1.0
  local singlepricewidth = 1.5
  local discountwidth = 1.5
  local sumwidth = 1.5
  if #self.allunits == 1 then
      unitwidth = 0
  end
  self:summarize()
  print('DISCOUNT', self.sums.discount)
  if self.sums.discount == 0 then
      discountwidth = 0
  end
  if not self.declaretax then
      taxwidth = 0
  end
  textwidth = fullwidth - countwidth - unitwidth - taxwidth - singlepricewidth - discountwidth - sumwidth
  context.starttabulate({
    string.format("|rg{%s}w(%scm)|lw(%scm)|lp(%scm)|rg{%s}w(%scm)|rg{%s}w(%scm)|rw(%scm)|rg{%s}w(%scm)|", sep, countwidth, unitwidth, textwidth, sep, taxwidth, sep, singlepricewidth, discountwidth, sep, sumwidth)
  })
  -- title
  if self.printtitle then
      context.NC(self.amountname[self.language])
      context.NC()
      context.NC()
      context.NC(self.taxname[self.language])
      context.NC(self.singlename[self.language])
      if self.printdiscounttitle then
        context.NC(self.discountname[self.language])
      else
        context.NC()
      end
      context.NC(self.sumname[self.language])
      context.NC()
      context.NR()
      context.HL()
  end
  -- items
  -- print into ConTeXt tabulate or table (old style)
  for _, item in pairs(self.items) do
    if item.typ == 'text' then
      context.NC()
      context.NC()
      context.NC(item.text)
      context.NC()
      context.NC()
      context.NC()
      context.NC()
      context.NR()
    elseif item.typ == 'lump' then
      context.NC()
      context.NC()
      context.NC(item.text)
      if self.declaretax then
        context.NC(self:fmt_percent(item.tax))
      else
        context.NC()
      end
      context.NC()
      context.NC(self:fmt_percent(item.discount, true))
      if self.printprice then
          context.NC(self:fmt_currency(item:sum()))
      else
          context.NC()
      end
      context.NC()
      context.NR()
    elseif item.typ == 'porto' then
      if self.printprice then
        context.NC()
        context.NC()
        context.NC(item.text)
        if self.declaretax then
          context.NC(self:fmt_percent(item.tax))
        else
          context.NC()
        end
        context.NC()
        context.NC(self:fmt_percent(item.discount, true))
        context.NC(self:fmt_currency(item:sum()))
        context.NC()
        context.NR()
      end
    else
      context.NC(self:fmt_number(item.count))
      context.NC(item.unit)
      context.NC(item.text)
      if self.declaretax then
        context.NC(self:fmt_percent(item.tax))
      else
        context.NC()
      end
      context.NC("à " .. self:fmt_currency(item.amount))
      context.NC(self:fmt_percent(item.discount, true))
      if self.printprice then
          context.NC(self:fmt_currency(item:sum()))
      else
          context.NC()
      end
      context.NC()
      context.NR()
    end
  end
  context.HL()
  if self.printprice then
      -- sums
      local endsum = self.sums.netto
      if self.printhoursum then
        context.NC(self:fmt_number(self.sums.hours))
        context.NC(self.timeunit)
      else
        context.NC()
        context.NC()
      end
      context.NC(self.sumname[self.language] .. ' ' .. self.taxmode)
      context.NC()
      context.NC()
      context.NC()
      context.NC(self.bold .. self:fmt_currency(endsum))
      context.NC()
      context.NR()
      if self.taxmode == 'netto' then
        endsum = self.sums.brutto
        local brutto = self.sums.netto
        for _, t in pairs(self.valid_tax) do
          if self.sums[t] > 0 then
            local tax = self.sums[t]
            brutto = brutto + tax
            context.NC()
            context.NC()
            context.NC(self.taxname[self.language])
            context.NC(self:fmt_percent(t))
            context.NC(self:fmt_currency(tax))
            context.NC()
            context.NC(self:fmt_currency(brutto))
            context.NC()
            context.NR()
          end
        end
        context.NC()
        context.NC()
        context.NC(self.sumname[self.language] .. ' brutto')
        context.NC()
        context.NC()
        context.NC()
        context.NC(self.bold .. self:fmt_currency(endsum))
        context.NC()
        context.NR()
      else
        for _, t in pairs(self.valid_tax) do
          if self.sums[t] > 0 then
            context.NC()
            context.NC()
            context.NC('inkl. ' .. self.taxname[self.language])
            context.NC(self:fmt_percent(t))
            context.NC(self:fmt_currency(self.sums[t]))
            context.NC()
            context.NC()
            context.NC()
            context.NR()
          end
        end
      end
      if self.currency2 ~= self.currency1 then
        local converted = endsum * self.exchangerate
        context.NC()
        context.NC()
        context.NC()
        context.NC()
        context.NC()
        context.NC()
        context.NC("(" .. self:fmt_currency(converted, self.currency2) .. ")")
        context.NC()
        context.NR()
      end
  end
  context.stoptabulate()
end
